<?php
// $Id: ffmpeg.inc,v 1.3.2.20 2008/12/21 19:03:08 jbrown Exp $

// Copyright 2007-2008 Jonathan Brown

define("OP_VIDEO_DEFAULT_FFMPEG_BINARY", 'ffmpeg');
define("OP_VIDEO_DEFAULT_QT_FASTSTART_BINARY", 'qt-faststart');


function _op_video_ffmpeg_exec($params, &$output, &$result, $mode = 0) {

  $ffmpeg = variable_get('op_video_ffmpeg_binary', OP_VIDEO_DEFAULT_FFMPEG_BINARY);
  $extra_params = variable_get('op_video_ffmpeg_extra_params', '');

  $command = array($ffmpeg, $params, $extra_params, '2>&1');

  _op_video_exec($command, $output, $result, $mode);
  _op_video_reset_cron();
}


function _op_video_qt_faststart_exec($params, &$output, &$result, $mode = 0) {

  $qt_faststart = variable_get('op_video_qt_faststart_binary', OP_VIDEO_DEFAULT_QT_FASTSTART_BINARY);

  $command = array($qt_faststart, $params, '2>&1');
  
  _op_video_exec($command, $output, $result, $mode);
  _op_video_reset_cron();
}


function _op_video_ffmpeg_version_info_fieldset($output) {
  array_shift($output);

  foreach($output as $i => $line) {
    $output[$i] = str_replace('  ', '&nbsp;&nbsp;', $line);
  }

  $output = implode("<br />\n", $output);
  
  $fieldset = array(
    '#type' => 'fieldset',
    '#title' => t('Version info'),
    '#collapsible' => TRUE,
    '#collapsed' => TRUE
  );

  $fieldset[] = array(
    '#value' => '<code>' . $output . '</code>'
  );
  
  return $fieldset;
}


function _op_video_ffmpeg_dot() {

  $url = base_path() . 'misc/menu-leaf.png';

  return array(
    'data' => '<img src="' . $url . '" />', 
    'style' => 'text-align: center;'
  );
}


function _op_video_ffmpeg_formats_fieldset(&$output) {
  $dot = _op_video_ffmpeg_dot();

  while(($line = array_shift($output))) {
    if($line == 'File formats:')
      break;
  }

  if(!$line)
    return;

  while(($line = array_shift($output)) != '') {
  
    preg_match('/ (.)(.) ([^ ]*) *(.*)/', $line, $matches);
    
    $important = in_array($matches[3], array('flv', 'mp4'));

    $rows[] = array(
      'data' => array(
        $matches[3],
        $matches[4],
        $matches[1] == 'D' ? $dot : '',
        $matches[2] == 'E' ? $dot : ''
      ),
      'style' => $important ? 'background-color: #425F95; color: white;' : NULL      
    );
  }

  $header = array(
    t('Name'), 
    t('Description'),
    array(
      'data' => t('Decode'), 
      'style' => 'text-align: center;'
    ),
    array(
      'data' => t('Encode'), 
      'style' => 'text-align: center;'
    ),
  );

  $fieldset = array(
    '#type' => 'fieldset',
    '#title' => t('File formats'),
    '#collapsible' => TRUE,
    '#collapsed' => TRUE
  );

  $fieldset[] = array(
    '#value' => theme('table', $header, $rows)
  );
  
  $note  = 'Note, the names of decoders and encoders do not always match, so there may be cases where the above table ';
  $note .= 'shows decode only or encode only entries even though both decoding and encoding are supported, ';
  $note .= 'e.g. mov / mp4 / 3gp.';

  $fieldset[] = array(
    '#value' => '<p>' . t($note) . '</p>'
  );
  
  return $fieldset;
}


function _op_video_ffmpeg_codecs_fieldset(&$output, &$codecs) {
  $dot = _op_video_ffmpeg_dot();

  $audio[] = array(array(
    'data' => '<b>' . t('Audio codecs') . '</b>', 
    'colspan' => '5'
  ));

  $video[] = array(array(
    'data' => '<b>' . t('Video codecs') . '</b>', 
    'colspan' => '5'
  ));

  while(($line = array_shift($output))) {
    if($line == 'Codecs:')
      break;
  }

  if(!$line)
    return;
    
  $codecs = array(
    'flv' => FALSE,
    'libmp3lame' => FALSE,
    'libx264' => FALSE,
    'libfaac' => FALSE
  );

  while(($line = array_shift($output)) != '') {
  
    preg_match('/ (.)(.)(.)... ([^ ]*) *(.*)/', $line, $matches);

    if(array_key_exists($matches[4], $codecs))
      $codecs[$matches[4]] = TRUE;
    
    $codec = array(
      '',
      $matches[4],
      $matches[5],
      $matches[1] == 'D' ? $dot : '',
      $matches[2] == 'E' ? $dot : ''
    );
    
    switch($matches[3]) {
    
      case 'A':
        $audio[] = array(
          'data' => $codec,
          'style' => $codecs[$matches[4]] ? 'background-color: #425F95; color: white;' : NULL
        );
        break;
        
      case 'V':
        $video[] = array(
          'data' => $codec,
          'style' => $codecs[$matches[4]] ? 'background-color: #425F95; color: white;' : NULL
        );
        break;
    }
  }
  
  $header = array(
    '',
    t('Name'), 
    t('Description'), 
    array(
      'data' => t('Decode'), 
      'style' => 'text-align: center;'
    ),
    array(
      'data' => t('Encode'), 
      'style' => 'text-align: center;'
    ),
  );

  $fieldset = array(
    '#type' => 'fieldset',
    '#title' => t('Codecs'),
    '#collapsible' => TRUE,
    '#collapsed' => TRUE
  );

  $fieldset[] = array(
    '#value' => theme('table', $header, array_merge($audio, $video))
  );

  $note  = 'Note, the names of decoders and encoders do not always match, so there may be cases where the above table ';
  $note .= 'shows decode only or encode only entries even though both decoding and encoding are supported, ';
  $note .= 'e.g. h264 / libx264.';

  $fieldset[] = array(
    '#value' => '<p>' . t($note) . '</p>'
  );

  return $fieldset;
}


function _op_video_ffmpeg_settings_fieldset($active) {

  $fieldset = array(
    '#type' => 'fieldset',
    '#title' => t('Local FFmpeg installation'),
    '#collapsible' => TRUE,
    '#collapsed' => !$active
  );

  $fieldset['op_video_ffmpeg_binary'] = array(
    '#type' => 'textfield',
    '#title' => t('FFmpeg binary'),
    '#description' => t('If the binary is not in the default path, you should enter the absolute path to it.'),
    '#default_value' => variable_get('op_video_ffmpeg_binary', OP_VIDEO_DEFAULT_FFMPEG_BINARY),
  );
  
  $fieldset['op_video_ffmpeg_extra_params'] = array(
    '#type' => 'textfield',
    '#title' => t('Extra parameters (advanced)'),
    '#description' => t('Any extra <a href="http://ffmpeg.mplayerhq.hu/ffmpeg-doc.html">parameters</a> you wish to have appended onto the FFmpeg command.'),
    '#default_value' => variable_get('op_video_ffmpeg_extra_params', ''),
  );
  
  $output = array();
  _op_video_ffmpeg_exec('-version', $output, $result);
  
  if($result == 0) {
    $fieldset[] = _op_video_ffmpeg_version_info_fieldset($output);

    $output = array();
    _op_video_ffmpeg_exec('-formats', $output, $result);

    $fieldset[] = _op_video_ffmpeg_formats_fieldset($output);
    $fieldset[] = _op_video_ffmpeg_codecs_fieldset($output, $codecs);
    
    if(!$codecs['libmp3lame']) {
      $fieldset[] = array(
        '#value' => '<p class="error">' . t('libmp3lame codec not found - FLV encoding will fail.') . '</p>'
      );
    }
    
    if(!$codecs['libx264']) {
      $fieldset[] = array(
        '#value' => '<p class="error">' . t('libx264 codec not found - MP4 encoding will fail.') . '</p>'
      );
    }
    
    if(!$codecs['libfaac']) {
      $fieldset[] = array(
        '#value' => '<p class="error">' . t('libfaac codec not found - MP4 encoding will fail.') . '</p>'
      );
    }
  }
  else {
    $fieldset[] = array(
      '#value' => '<p class="error">' . t('FFmpeg not available - transcoding will fail.') . '</p>'
    );
  }
  
  $fieldset['op_video_qt_faststart_binary'] = array(
    '#type' => 'textfield',
    '#title' => t('qt-faststart binary'),
    '#description' => t('If the binary is not in the default path, you should enter the absolute path to it.'),
    '#default_value' => variable_get('op_video_qt_faststart_binary', OP_VIDEO_DEFAULT_QT_FASTSTART_BINARY),
  );
  
  $output = array();
  _op_video_qt_faststart_exec(NULL, $output, $result);
  
  if($result != 0) {
    $fieldset[] = array(
      '#value' => '<p class="error">' . t('qt-faststart not available - MP4 encoding will fail.') . '</p>'
    );
  }

  return $fieldset;
}


function _op_video_ffmpeg_get_info($source_file) {
  
  $params = '-i ' . escapeshellarg($source_file['filepath']);
  $output = array();
  _op_video_ffmpeg_exec($params, $output, $result);
  array_shift($output);
  $output = implode("\n", $output);
  
  if(preg_match('/Duration: *(..):(..):(..\..*), /', $output, $matches))
    $info['duration'] = ($matches[1] * 60 + $matches[2]) * 60 + $matches[3];
  else
    return NULL;
    
  if(preg_match('/Audio: (.*)/', $output, $matches)) {
    $parts = explode(', ', $matches[1]);
    
    $info['audio_codec'] = strtok(array_shift($parts), ' ');
    $info['audio_sample_rate'] = strtok(array_shift($parts), ' ');
    $info['audio_channels'] = array_shift($parts);
    
    switch($info['audio_channels']) {
      
      case 'mono':
        $info['audio_channels'] = 1;
        break;

      case 'stereo':
        $info['audio_channels'] = 2;
        break;
    }
    
    if(count($parts) > 1)
      array_shift($parts);
      
    $info['audio_bitrate'] = strtok(array_shift($parts), ' ');
  }
  else {
    $info['audio_channels'] = 0;
  }

  if(preg_match('/Video: (.*)/', $output, $matches)) {
  
    $parts = explode(', ', $matches[1]);
    
    $info['video_codec'] = array_shift($parts);
    array_shift($parts);
    $dims = explode('x', array_shift($parts));
    $info['width'] = $dims[0];
    $info['height'] = strtok($dims[1], ' ');
    
    if(count($parts) > 1)
      $info['video_bitrate'] = strtok(array_shift($parts), ' ');
    else
      $info['video_bitrate'] = $source_file['filesize'] * 8 / 1000 / $info['duration'] - $info['audio_bitrate'];
      
    $info['frame_rate'] = strtok(array_shift($parts), ' ');
  }

  $info['source_file'] = $source_file;
  
  return $info;
}


function _op_video_ffmpeg_create_png($video_id, $info, $i, &$output) {
  
  $params[] = '-v 0 -y -i ' . escapeshellarg($info['source_file']['filepath']);

  $out_file = _op_video_temp_path() . '/' . $video_id . '_raw_image_' . $i . '.png';

  if($i == 2)
    $params[] = '-ss ' . $info['duration'] / 3;

  $params[] = '-vcodec png';
  $params[] = '-vframes 1';
  $params[] = '-f rawvideo';
  $params[] = '-s ' . ((int)($info['width'] / 2)) * 2 . 'x' . ((int)($info['height'] / 2)) * 2;
  $params[] = '-an';
  $params[] = $out_file;

  _op_video_ffmpeg_exec($params, $output, $result, OP_VIDEO_EXEC_MODE_TIME | OP_VIDEO_EXEC_MODE_ECHO);

  if($result == 0) {
    file_move($out_file, _op_video_transcoded_path(), FILE_EXISTS_REPLACE);
    return _op_video_insert_filepath($out_file);
  }
  else {
    file_delete($out_file);
    return FALSE;
  }
}


function _op_video_ffmpeg_create_flv($video_id, $info, $dirty_params, &$output) {

  $params[] = '-v 0 -y -i ' . escapeshellarg($info['source_file']['filepath']);

  if($dirty_params['max_duration'] != 0 && 
    $info['duration'] > $dirty_params['max_duration'])
  {
    $params[] = '-t ' . $dirty_params['max_duration'];
  }

  if($info['video_codec'] == 'flv' &&
    $info['video_bitrate'] <= $dirty_params['video_bitrate'] &&
    $info['width'] <= $dirty_params['width'] &&
    $info['frame_rate'] <= $dirty_params['frame_rate'])
  {
    $params[] = '-vcodec copy';
    $passthrough = 1;
  }
  else {
    $params[] = '-vcodec flv';
    $params[] = '-b ' . $dirty_params['video_bitrate'] . 'k';

    if($info['frame_rate'] > $dirty_params['frame_rate']) {
      $frame_rate = $dirty_params['frame_rate'];
      $params[] = '-r ' . $frame_rate;
    }
    else {
      $frame_rate = $info['frame_rate'];
    }
    
    $params[] = '-g ' . (int)round($frame_rate * $dirty_params['key_interval']);

    if($info['width'] > $dirty_params['width']) {

      $width = $dirty_params['width'];
      $height = (int)round(($dirty_params['width'] * $info['height']) / $info['width']);
    }
    else {
      $width = $info['width'];
      $height = $info['height'];
    }
    
    $params[] = '-s ' . ((int)($width / 2)) * 2 . 'x' . ((int)($height / 2)) * 2;
  }
  
  $channels = min($info['audio_channels'], $dirty_params['audio_channels']);

  if($channels == 0) {
    $params[] = '-an';
  }
  else if($info['audio_codec'] == 'mp3' &&
    in_array($info['audio_sample_rate'], array(11025, 22050, 44100)) &&
    $info['audio_bitrate'] <= $dirty_params['audio_bitrate'] &&
    $info['audio_channels'] <= $dirty_params['audio_channels'] &&
    $info['audio_sample_rate'] <= $dirty_params['audio_sample_rate'])
  {
    $params[] = '-acodec copy';
  }
  else {
    $params[] = '-acodec libmp3lame';
    $params[] = '-ab ' . $dirty_params['audio_bitrate'] . 'k';
    $params[] = '-ac ' . $channels;
    $params[] = '-ar ' . $dirty_params['audio_sample_rate'];
  }
  
  $out_file = _op_video_temp_path() . '/' . $video_id . '.flv';

  if($passthrough) {
    _op_video_ffmpeg_exec(array($params, $out_file), $output, $result, OP_VIDEO_EXEC_MODE_TIME | OP_VIDEO_EXEC_MODE_ECHO);
  }
  else {
    $log_file = _op_video_temp_path() . '/2pass-' . $video_id;

    _op_video_ffmpeg_exec(array($params, '-pass 1 -passlogfile ' . $log_file, $out_file), 
      $output, $result, OP_VIDEO_EXEC_MODE_TIME | OP_VIDEO_EXEC_MODE_ECHO);

    if($result == 0) {
      _op_video_ffmpeg_exec(array($params, '-pass 2 -passlogfile ' . $log_file, $out_file), 
        $output, $result, OP_VIDEO_EXEC_MODE_TIME | OP_VIDEO_EXEC_MODE_ECHO);

      if($result != 0) {  // dual pass failed - try single pass
        _op_video_ffmpeg_exec(array($params, $out_file), 
          $output, $result, OP_VIDEO_EXEC_MODE_TIME | OP_VIDEO_EXEC_MODE_ECHO);
      }
    }

    file_delete($log_file . '-0.log');
  }
  
  if($result == 0) {
    file_move($out_file, _op_video_transcoded_path(), FILE_EXISTS_REPLACE);
    return _op_video_insert_filepath($out_file);
  }
  else {
    file_delete($out_file);
    return FALSE;
  }
}


function _op_video_ffmpeg_create_mp4($video_id, $info, $dirty_params, &$output) {
  $original_dir = getcwd();
  
  $params[] = '-v 0 -y -i ' . escapeshellarg($original_dir . '/' . $info['source_file']['filepath']);

  if($dirty_params['max_duration'] != 0 && 
    $info['duration'] > $dirty_params['max_duration'])
  {
    $params[] = '-t ' . $dirty_params['max_duration'];
  }

  if($info['video_codec'] == 'h264' && 
    $info['video_bitrate'] <= $dirty_params['video_bitrate'] &&
    $info['width'] <= $dirty_params['width'] &&
    $info['frame_rate'] <= $dirty_params['frame_rate'])
  {
    $params[] = '-vcodec copy';
    $passthrough = 1;
  }
  else {
    $params[] = '-vcodec libx264';
    $params[] = '-b ' . $dirty_params['video_bitrate'] . 'k';

    if($info['frame_rate'] > $dirty_params['frame_rate']) {
      $frame_rate = $dirty_params['frame_rate'];
      $params[] = '-r ' . $frame_rate;
    }
    else {
      $frame_rate = $info['frame_rate'];
    }
    
    $params[] = '-g ' . (int)round($frame_rate * $dirty_params['key_interval']);

    if($info['width'] > $dirty_params['width']) {

      $width = $dirty_params['width'];
      $height = (int)round(($dirty_params['width'] * $info['height']) / $info['width']);
    }
    else {
      $width = $info['width'];
      $height = $info['height'];
    }
    
    $params[] = '-s ' . ((int)($width / 2)) * 2 . 'x' . ((int)($height / 2)) * 2;
  }
  
  $channels = min($info['audio_channels'], $dirty_params['audio_channels']);

  if($channels == 0) {
    $params[] = '-an';
  }
  else if($info['audio_codec'] == 'aac' &&
    $info['audio_bitrate'] <= $dirty_params['audio_bitrate'] &&
    $info['audio_channels'] <= $dirty_params['audio_channels'] &&
    $info['audio_sample_rate'] <= $dirty_params['audio_sample_rate'])
  {
    $params[] = '-acodec copy';
  }
  else {
    $params[] = '-acodec libfaac';
    $params[] = '-ab ' . $dirty_params['audio_bitrate'] . 'k';
    $params[] = '-ac ' . $channels;
    $params[] = '-ar ' . $dirty_params['audio_sample_rate'];
  }
  
  $pre_file = _op_video_temp_path() . '/' . $video_id . '-pre.mp4';

  if($passthrough) {
    _op_video_ffmpeg_exec(array($params, $pre_file), $output, $result, 
      OP_VIDEO_EXEC_MODE_TIME | OP_VIDEO_EXEC_MODE_ECHO);
  }
  else {
    chdir(_op_video_temp_path());   // this is necessary because libx264 dumps its log file in the current directory,
                                    // which usually isn't writable

    $log_file = '2pass-' . $video_id;
    
    _op_video_ffmpeg_exec(array($params, '-pass 1 -passlogfile ' . $log_file, $original_dir . '/' . $pre_file), 
      $output, $result, OP_VIDEO_EXEC_MODE_TIME | OP_VIDEO_EXEC_MODE_ECHO);

    if($result == 0) {
      _op_video_ffmpeg_exec(array($params, '-pass 2 -passlogfile ' . $log_file, $original_dir . '/' . $pre_file), 
        $output, $result, OP_VIDEO_EXEC_MODE_TIME | OP_VIDEO_EXEC_MODE_ECHO);

      if($result != 0) {  // dual pass failed - try single pass
        _op_video_ffmpeg_exec(array($params, $original_dir . '/' . $pre_file), 
          $output, $result, OP_VIDEO_EXEC_MODE_TIME | OP_VIDEO_EXEC_MODE_ECHO);
      }
    }

    file_delete($log_file . '-0.log');
    file_delete('x264_2pass.log');

    chdir($original_dir);
  }
  
  if($result != 0) {
    file_delete($pre_file);
    return FALSE;
  }

  $out_file = _op_video_temp_path() . '/' . $video_id . '.mp4';
  
  _op_video_qt_faststart_exec(array($pre_file, $out_file), 
    $output, $result, OP_VIDEO_EXEC_MODE_TIME | OP_VIDEO_EXEC_MODE_ECHO);

  file_delete($pre_file);
  
  if($result == 0) {
    file_move($out_file, _op_video_transcoded_path(), FILE_EXISTS_REPLACE);
    return _op_video_insert_filepath($out_file);
  }
  else {
    file_delete($out_file);
    return FALSE;
  }
}


function _op_video_transcode_ffmpeg($video) {

  $info = _op_video_ffmpeg_get_info($video['source_file']);
  
  if($info) {
    switch($video['dirty_params']['format']) {
    
      case OP_VIDEO_FORMAT_FLV:
        $flv_file_id = _op_video_ffmpeg_create_flv($video['video_id'], $info, $video['dirty_params'], $output);
        break;
    
      case OP_VIDEO_FORMAT_MP4:
        $flv_file_id = _op_video_ffmpeg_create_mp4($video['video_id'], $info, $video['dirty_params'], $output);
        break;
    }
    
    if($flv_file_id) {
      $start_image_file_id = _op_video_ffmpeg_create_png($video['video_id'], $info, 1, $output);
      $third_image_file_id = _op_video_ffmpeg_create_png($video['video_id'], $info, 2, $output);

      $current_params_id = _op_video_transcoding_params_insert($video['dirty_params']);

      db_query("
        UPDATE {op_videos}
        SET duration = %d,
          flv_file_id = %d,
          start_image_file_id = %d,
          third_image_file_id = %d,
          current_params_id = %d,
          status = 'idle',
          error = 'none'
        WHERE video_id = %d
      ",
        (int)ceil($info['duration']),
        $flv_file_id,
        $start_image_file_id,
        $third_image_file_id,
        $current_params_id,
        $video['video_id']
      );

      _op_video_transcoding_completed($video);
      return;
    }
  }

  _op_video_set_transcoding_failed($video['video_id']);
}

